/**
 * @author Luc Yriarte
 * 
 * Licence : BSD http://www.opensource.org/licenses/bsd-license.php
 *  
 */
package basicSim;
/* Java 2
import java.util.ArrayList;
import java.util.StringTokenizer;
*/
/**
 * Mesh wireframe model. Edges are represented by the indexes of both ends
 * in the vertex table.
 */
public class Mesh {
	public int nVertex, nEdge;
	/** vertex table */
	public Vector3D[] vertex;
	/** an edge is a pair of indexes in the vertex table */
	public int[][] edge;
	/** wireframe screen coordinates */
	int focal, screenX, screenY;
	/** vertices xy translated in screen coordinates */
	protected int scrX[], scrY[];
	/** wireframe lines as an array of edges quintuplets <x1 y1 x2 y2 (z1+z2)/2> */
	protected int lines[][];

	public Mesh() {
		nVertex = nEdge = 0;
		vertex = null;
		edge = null;
		lines = null;
		scrX = scrY = null;
	}

	public Mesh(int nV, int nE, Vector3D[] v, int[][] e) {
		nVertex = nV;
		nEdge = nE;
		vertex = v;
		edge = e;
		scrX = new int[nVertex];
		scrY = new int[nVertex];
		lines = new int[nEdge][5];		
	}

/* Java 2
	public Mesh(String v, String e) {
		StringTokenizer st;
		String tok;
		ArrayList values;
		double x, y, z;
		int i;
		values = new ArrayList();
		st = new StringTokenizer(v, "[]()<>, \\s");
		while (st.hasMoreTokens()) {
			tok = st.nextToken();
			x = (new Double(tok)).doubleValue();
			tok = st.nextToken();
			y = (new Double(tok)).doubleValue();
			tok = st.nextToken();
			z = (new Double(tok)).doubleValue();
			values.add(new Vector3D(x, y, z));
		}
		nVertex = values.size();
		vertex = new Matrix[nVertex];
		for (i = 0; i < nVertex; i++) {
			vertex[i] = (Matrix) values.get(i);
		}
		values = new ArrayList();
		st = new StringTokenizer(e, "[]()<>, \\s");
		while (st.hasMoreTokens()) {
			tok = st.nextToken();
			values.add(new Integer(tok));
			tok = st.nextToken();
			values.add(new Integer(tok));
		}
		nEdge = values.size() / 2;
		edge = new int[nEdge][2];
		for (i = 0; i < nEdge; i++) {
			edge[i][0] = ((Integer) values.get(i * 2)).intValue();
			edge[i][1] = ((Integer) values.get(i * 2 + 1)).intValue();
		}
		scrX = new int[nVertex];
		scrY = new int[nVertex];
		lines = new int[nEdge][5];		
	}
*/
	/**
	 * Creates a new mesh with the transformed vertices and this meshes
	 * edges.
	 */
	public Mesh transform(Matrix trans) {
		if (nVertex == 0)
			return this;
		int i;
		Vector3D[] transVertex = new Vector3D[nVertex];
		for (i = 0; i < nVertex; i++) {
			transVertex[i] = Vector3D.MatrixAsVector3D(trans.mul(vertex[i]));
		}
		return new Mesh(nVertex, nEdge, transVertex, edge);
	}

	/**
	 * Transform this mesh
	 */
	public Mesh transformThis(Matrix trans) {
		if (nVertex == 0)
			return this;
		int i;
		for (i = 0; i < nVertex; i++) {
			vertex[i].transformThis(trans);
		}
		return this;
	}

	/**
	 * Converts the XYZ coordinates of each vertex in flat xy coordinates
	 *  
	 * @param foc: focal focal distance
	 * @param sX: screenX screen dimension x
	 * @param sY: screenY screen dimension y
	 * @return this
	 */
	public Mesh updateWireframe(int foc, int sX, int sY) {
		if (nEdge == 0)
			return null;
		int i, x0, y0;
		double z;
		focal = foc;
		screenX = sX;
		screenY = sY;
		// screen center
		x0 = screenX / 2;
		y0 = screenY / 2;
		/** wireframe lines */
//		int lines[][];
		/** vertices translated screen coordinates */
//		int scrX[], scrY[];
//		scrX = new int[nVertex];
//		scrY = new int[nVertex];
//		lines = new int[nEdge][5];
		// Translate vertices to screen coordinates
		for (i = 0; i < nVertex; i++) {
			z = vertex[i].cell[2][0] > 1 ? vertex[i].cell[2][0] : 1;
			scrX[i] = (int) (x0 + vertex[i].cell[0][0] * focal / z);
			scrY[i] = (int) (y0 - vertex[i].cell[1][0] * focal / z);
		}
		// Create lines from mesh edges
		for (i = 0; i < nEdge; i++) {
			lines[i][0] = scrX[edge[i][0]];
			lines[i][1] = scrY[edge[i][0]];
			lines[i][2] = scrX[edge[i][1]];
			lines[i][3] = scrY[edge[i][1]];
			lines[i][4] = ((int) (vertex[edge[i][0]].cell[2][0] + vertex[edge[i][1]].cell[2][0])) >> 1;
		}
		return this;
	}
	
	public Mesh setWireframe(Mesh aMesh) {
		this.screenX = aMesh.screenX;
		this.screenY = aMesh.screenY;
		this.focal = aMesh.focal;
		this.scrX = aMesh.scrX;
		this.scrY = aMesh.scrY;
		this.lines = aMesh.lines;
		return this;
	}
	
	/**
	 * Selects the first vertex where xy coordinates in the mesh
	 * wireframe are closer than r to xy parameters.
	 * @param x
	 * @param y
	 * @param r
	 * @return index of the selected vertex, -1 if none
	 */
	public int selectVertex(int x, int y, int r) {
		int xmr,xpr,ymr,ypr;
		int i;
		i=-1;
		xmr=x-r;xpr=x+r;ymr=y-r;ypr=y+r;
		for (i = 0; i < nVertex; i++) {
			if (scrX[i] >= xmr && scrX[i] <= xpr && scrY[i] >= ymr && scrY[i] <= ypr)
				break;
		}
		if (i==nVertex)
			i=-1;
// System.out.println("select vertex [" + xmr + ".." + xpr + "][" + ymr + ".." + ypr + "] :" + i);		
		return i;
	}
	
}
